@@ -710,8 +710,8 @@ static const GDBusMethodTable agent_methods[] = {
 	{ },
 };
 
-static int agent_register_return(DBusMessageIter *iter, const char *error,
-		void *user_data)
+static int agent_register_return(DBusMessageIter *iter, int errnum,
+				 const char *error, void *user_data)
 {
 	DBusConnection *connection = user_data;
 
@@ -768,8 +768,8 @@ int __connmanctl_agent_register(DBusConnection *connection)
 	return result;
 }
 
-static int agent_unregister_return(DBusMessageIter *iter, const char *error,
-		void *user_data)
+static int agent_unregister_return(DBusMessageIter *iter, int errnum,
+				   const char *error, void *user_data)
 {
 	if (error) {
 		fprintf(stderr, "Error unregistering Agent: %s\n", error);
@@ -819,8 +819,8 @@ static const GDBusMethodTable vpn_agent_methods[] = {
 	{ },
 };
 
-static int vpn_agent_register_return(DBusMessageIter *iter, const char *error,
-		void *user_data)
+static int vpn_agent_register_return(DBusMessageIter *iter, int errnum,
+				     const char *error,	void *user_data)
 {
 	DBusConnection *connection = user_data;
 
@@ -872,7 +872,7 @@ int __connmanctl_vpn_agent_register(DBusConnection *connection)
 	return result;
 }
 
-static int vpn_agent_unregister_return(DBusMessageIter *iter,
+static int vpn_agent_unregister_return(DBusMessageIter *iter, int errnum,
 		const char *error, void *user_data)
 {
 	if (error) {
